(in-package #:cl-user)
(defpackage #:utils
  (:use #:cl)
  (:import-from #:com.inuoe.jzon
                #:parse
                #:stringify)
  (:import-from #:cl-fad
                #:pathname-directory-pathname)
  (:import-from #:local-time
                #:format-timestring
                #:now)
  (:import-from #:log4cl
                #:warn
                #:debug
                #:info
                #:error)
  (:local-nicknames (#:jzon #:com.inuoe.jzon))
  (:shadow #:dexador
	   #:request)
  (:export #:check-config-file
           #:make-config-file
           #:read-config-file
           #:send-notification
           #:update-ip-addresses)
  (:documentation "Back-end for ddns-updater."))
(in-package #:utils)

(defun add-filename-safe-timestamp (text)
  "Adds a filename-safe time stamp (when the function is called) to `TEXT'.
The intended use for this function is to create time stamped versions of
filenames, which can then be stored as back-up copies.

The format takes the following shape: YYYYMMDD_HHMMSS_filename.ext

Example:
========

'config.json' becomes '20231217_14205_config.json'."
  (format nil "~a_~a"
          (local-time:format-timestring
           nil
           (local-time:now)
           :format '(:year :month :day "_" :hour :min :sec))
          text))

(defun backup-config-file (path)
  "Creates a back-up file for the programs config. file.
The back-up file add a filename safe time stamp to the config. file and is
stored in the same directory as the previous config. file, which is specified at
`PATH'.

Example:
========

 ~/Desktop/basic.json will be coped and stored as
 ~/Desktop/20231217_14298_basic.json"
  (uiop:copy-file
   path
   (format nil "~a~a"
           (cl-fad:pathname-directory-pathname
            (uiop:native-namestring path))
           (add-filename-safe-timestamp (file-namestring path)))))

(defun get-accounts (conf-data path)
  "Gets the account data for the specified DNSimple account."
  (handler-case
      (progn
        (log:debug "Getting accounts...")
        (let* ((token (gethash "token" conf-data))
               (data 
                 (dexador:request
                  "https://api.dnsimple.com/v2/accounts"
                  :method :get
                  :use-connection-pool nil
                  :keep-alive nil
                  :verbose nil
                  :headers `(("Accept" . "application/json")
                             ("Authorization" . ,(format nil "Bearer ~a" token))))))
          (with-open-file (stream path
                                  :direction :output
                                  :if-exists :supersede)
            (jzon:stringify (jzon:parse data) :pretty t :stream stream))))
    (error (e)
      (log:error "~a" e))))

(defun get-ip-address ()
  "Get the public IP address of the server."
  (handler-case
      (progn
        (log:debug "Getting public IP address of server...")
        (string-trim
         '(#\Space #\e #\t #\m #\Newline)
         (dexador:request
          "https://icanhazip.com"
          :method :get
          :use-connection-pool nil
          :keep-alive nil
          :verbose nil)))
    (error (e)
      (log:error "~a" e))))

(defun get-domain-zones (conf-data path)
  "Gets the zones (domains) attached account and writes to JSON file to `PATH'."
  (handler-case
      (progn
        (log:debug "Getting (sub-domain) zones...")
        (let* ((account-id (gethash "account-id" conf-data))
               (token (gethash "token" conf-data))
               (url (format nil "https://api.dnsimple.com/v2/~a/zones" account-id))
               (data
                 (dexador:request
                  url
                  :method :get
                  :use-connection-pool nil
                  :keep-alive nil
                  :verbose nil
                  :headers `(("Content-Type" . "application/json")
		             ("Authorization" . ,(format nil "Bearer ~a" token))))))
          (with-open-file (stream path
                                  :direction :output
                                  :if-exists :supersede)
            (jzon:stringify (jzon:parse data) :pretty t :stream stream))))
    (error (e)
      (log:error "~a" e))))

(defun get-records-in-zone (conf-data path)
  "Creates a JSON file,listing the ID's (sub-domain) in a zone (top-level domain).
The Id's are used to update the IP address for a given (sub)domain name.  So,
43482356 is used to identify test.your-domain.com when using DNS service's REST
API. An example of what to look for in the JSON file produced by this function
is as follows,

- id: 43480936 (add this to your configuration file)
- zone_id: your-domain.com
- name: test"
  (handler-case
      (progn
        (log:debug "Getting DNS zones...")
        (let* ((account-id (gethash "account-id" conf-data))
               (zone (gethash "zone-id" conf-data))
               (token (gethash "token" conf-data))
               (url
                 (format nil "https://api.dnsimple.com/v2/~a/zones/~a/records"
                         account-id zone))
               (data
                 (dexador:request
                  url
                  :method :get
                  :use-connection-pool nil
                  :keep-alive nil
                  :verbose nil
                  :headers `(("Content-Type" . "application/json")
		             ("Authorization" . ,(format nil "Bearer ~a" token))))))
          (with-open-file (stream path
                                  :direction :output
                                  :if-exists :supersede)
            (jzon:stringify (jzon:parse data) :pretty t :stream stream))))
    (error (e)
      (log:error "~a" e))))

(defun check-config-file (path)
  "Checks to see if each key is present in the configuration file.
Prints a warning message to the console if a particular part of the JSON object
is missing."
  (handler-case
      (let ((conf-data (read-config-file path))
            (error-counter 0))
        (format t "Checking schema for ~a ~&" path)
        (when (null (gethash "token" conf-data))
          (progn
            (incf error-counter)
            (format t "[token]             DNSimple API Token is missing.~&")))
        (when (null (gethash "server-name" conf-data))
          (progn
            (format t "[server-name]       Server name is missing.~&")
            (incf error-counter)))
        (when (null (gethash "account-id" conf-data))
          (progn
            (format t "[account-id]        DNSimple Account Id. is missing.~&")
            (incf error-counter)))
        (when (null (gethash "record-ids" conf-data))
          (progn
            (format t "[record-ids]        DNSimple Record Id's aray are missing.~&")
            (incf error-counter)))
        (when (null (gethash "telegram-token" conf-data))
          (progn
            (format t "[telegram-token]    Telegram API token is missing.~&")
            (incf error-counter)))
        (when (null (gethash "chat-id" conf-data))
          (progn
            (format t "[chat-id]           Telegram Chat Id. is missing.~&")
            (incf error-counter)))
        (when (null (gethash "notify-on-success" conf-data))
          (progn
            (format t "[notify-on-success] Notify on Success option is missing.~&")
            (incf error-counter)))
        (when (null (gethash "notify-on-fail" conf-data))
          (progn
            (format t "[notify-on-fail]    Notify on Fail option is missing.~&")
            (incf error-counter)))
        (when (null (gethash "last-recorded-ip" conf-data))
          (progn
            (format t "[last-recorded-ip]  Last Recorded IP address is missing.~&")
            (incf error-counter)))
        (format t "Check complete with ~d errors found.~&" error-counter))
    (error (e)
      (log:error "~a" e))))

(defun make-config-file (path)
  "Creates a templated config. file, which the user can then fill out."
  (handler-case
      (progn
        (log:debug "Making config. file...")
        (let ((ht (make-hash-table :test 'equal))
              (native-path (uiop:native-namestring path)))
          ;; DNS
          (setf (gethash "token" ht) "your_oath_token")
          (setf (gethash "server-name" ht) "your_sever_name")
          (setf (gethash "account-id" ht) "your_account_id")
          (setf (gethash "record-ids" ht)
                (make-array 2 :initial-element "your_domain_name_record_id"))
          ;; Telegram
          (setf (gethash "telegram-token" ht) "your_telegram_token")
          (setf (gethash "chat-id" ht) "your_chat_id")
          (setf (gethash "notify-on-success" ht) "false")
          (setf (gethash "notify-on-fail" ht) "true")
          (setf (gethash "last-recorded-ip" ht) "0.0.0.0")
          (when (uiop:file-exists-p native-path)
            (log:warn "Config. file already exists, back-up created.")
            (backup-config-file native-path))
          (with-open-file (stream native-path
                                  :direction :output
                                  :if-exists :supersede)
            (com.inuoe.jzon:stringify ht :pretty t :stream stream))))
    (error (e)
      (log:error "~a" e))))

(defun read-config-file (path)
  "Reads the config. file used by this project.."
  (handler-case
      (let ((native-path (uiop:native-namestring path)))
        (log:debug "Reading config. file...")
        (let ((data (uiop:read-file-string native-path)))
          (com.inuoe.jzon:parse data)))
    (error (e)
      ;; Using this error case to catch 'file not found' errors, instead of
      ;; adding a check above in the `HANDLER-CASE'. It's not as explicit when
      ;; reading the code but the program catches the error, reports it to the
      ;; user and exits out of the program without being overly dramatic.
      (log:error "~a" e))))

(defun send-notification (api-token chat-id message)
  "Sends a `MESSAGE', using `API-TOKEN' and `CHAT-ID' provided by Telegram."
  (handler-case
      (progn
        (log:debug "Sending message to Telegram...")
        (dexador:request
         (format nil "https://api.telegram.org/bot~a/sendMessage" api-token)
         :method :post
         :use-connection-pool nil
         :keep-alive nil
         :verbose nil
         :headers '(("Content-Type" . "application/json"))
         :content
         (format nil "{ \"chat_id\": \"~a\", \"text\": \"~a\"}" chat-id message)))
    (error (e)
      (log:error "~a" e))))

(defun update-ip-addresses (conf-data conf-path)
  "Updates IP address' of sub-domains (record id) attached to same top-level domain (zone)."
  (handler-case
      (let ((old-ip (gethash "last-recorded-ip" conf-data))
            (new-ip (get-ip-address)))
        (unless (string= old-ip new-ip) ; Ignore unless they don't match.
          (map 'vector
               (lambda (sub-domain-id)
                 (log:debug "Updating: ~a~%" sub-domain-id)
                 (dexador:request
                  (format nil "https://api.dnsimple.com/v2/~a/zones/~a/records/~a"
                          (gethash "account-id" conf-data)
                          (gethash "zone-id" conf-data)
                          sub-domain-id)
                  :method :patch
                  :use-connection-pool t
                  :keep-alive t
                  :verbose nil
                  :headers `(("Content-Type" . "application/json")
		             ("Authorization" . ,(format nil "Bearer ~a"
                                                         (gethash "token" conf-data))))
                  :content (format nil "{ \"content\": \"~a\" }" new-ip)))
               (gethash "record-ids" conf-data))
          (when (string= "true" (gethash "notify-on-success" conf-data))
            (send-notification
             (gethash "telegram-token" conf-data)
             (gethash "chat-id" conf-data)
             (format nil "[🟩 SUCCESS] Updated IP address' on ~a for ~a."
                     (gethash "server-name" conf-data)
                     (gethash "zone-id" conf-data))))
          (update-conf-file-to-latest-ip conf-data new-ip conf-path)))
    (error (e)
      (log:error "~a" e)
      (when (and (gethash "telegram-token" conf-data)
                 (gethash "chat-id" conf-data))
        (send-notification
         (gethash "telegram-token" conf-data)
         (gethash "chat-id" conf-data)
         (format nil "[🔴 FAIL] Unable to update IP address' on ~a for ~a."
                 (gethash "server-name" conf-data)
                 (gethash "zone-id" conf-data)))))))

(defun update-conf-file-to-latest-ip (conf-data new-ip conf-path)
  "Changes the 'last-recorded-ip' property in the configuration file."
  (handler-case
      (progn
        (log:debug "Updating config. file to latest IP address...")
        (setf (gethash "last-recorded-ip" conf-data) new-ip)
        (with-open-file (stream (uiop:native-namestring conf-path)
                                :direction :output
                                :if-exists :supersede)
          (jzon:stringify conf-data :pretty t :stream stream)))
    (error (e)
      (log:error "~a" e))))
